import numpy as np
from scipy.special import erf
from IPython.display import Math, HTML, display

from . import ctmc
from . import fret


valid_model_names = ('gaussian', 'wlc', 'gaussian_chain', 'radial_gaussian')


def assert_valid_model_name(name):
    name = name.lower()
    if name not in valid_model_names:
        raise TypeError(f'Distance model name "{name}" not recognized.'
                        f'Valid model names are:\n {valid_model_names})')


def distribution(params):
    name = params['name'].lower()
    assert_valid_model_name(name)
    model = {'gaussian': Gaussian, 'radial_gaussian': RadialGaussian,
             'gaussian_chain': GaussianChain, 'wlc': WormLikeChain}
    return model[name](params)


def _check_k_s(k_s, num_states):
    """Check consistency of k_s parameter in case of more than 1 state.
    Returns k_s converted to a float-array.
    """
    if num_states == 1:
        return
    if k_s is None:
        msg = (f'Missing k_s parameters, necessary in a '
               f'{num_states}-states model.')
        raise ValueError(msg)
    k_s = np.asfarray(k_s)
    if k_s.ndim == 1 and num_states == 2:
        if np.size(k_s) != 2:
            msg = f'"k_s" (={k_s}) should be a 2-element 1d array or an 2x2 array.'
            raise TypeError(msg)
    else:
        if k_s.shape != (num_states, num_states):
            msg = f'"k_s" (={k_s}) should be an {num_states}x{num_states} array.'
            raise TypeError(msg)
    return k_s


class BaseDistribution:
    """Base distance distribution to be subclassed.

    Sub-classes must define the attributes `param_names` (tuple of parameters names)
    and `_pdf_latex` (a string with a LaTeX representation of the PDF).
    They should also define the static method `pdf` (a function defining the PDF),
    and the methods `r_max` (returning the r max value) and `_html_params` (returning
    a string with HTML representation of paramenters).
    """
    def __init__(self, params):
        self.name = params['name']
        self.params = {k: np.asfarray(v) for k, v in params.items()
                       if k in self.param_names}
        self.num_states = self._get_num_states()
        self._check_params(params)
        if self.num_states > 1:
            self._setup_multi_state(params.get('k_s'))

    def _setup_multi_state(self, k_s):
        self.k_s = k_s = _check_k_s(k_s, self.num_states)
        if k_s.ndim == 1:
            self.fract = np.array([k_s[1] / (k_s.sum()), k_s[0] / (k_s.sum())])
        else:
            self.fract = ctmc.CT_stationary_distribution(k_s)

    def _check_params(self, params):
        for param_name in self.param_names:
            if param_name not in params:
                raise TypeError(f'Missing parameter {param_name} in model "{self.name}".')
        for k, v in self.params.items():
            if np.size(v) != self.num_states:
                msg = (f'Parameter `{k}` ({v}) must have the same '
                       f'size as the number of states ({self.num_states}).')
                raise TypeError(msg)

    def _get_num_states(self):
        for k, v in self.params.items():
            num_states = np.size(v)
            break
        return num_states

    def get_r_axis(self, dr, n=5):
        r_max = self.r_max(n=n)
        start = np.min(self.params.get('offset', 0))
        return np.arange(start, r_max + dr, dr)

    def get_pdf(self, dr=0.01, n=5):
        r = self.get_r_axis(dr, n=n)
        if self.num_states == 1:
            y = self.pdf(r, **self.params)
        else:
            y = np.zeros_like(r)
            for i in range(self.num_states):
                p = {k: v[i] for k, v in self.params.items()}
                y += self.fract[i] * self.pdf(r, **p)
        return r, y

    def _ipython_display_(self):
        state = '1-state' if self.num_states == 1 else f'{self.num_states}-states'
        display(HTML(f'<h3>Distance Model: "{self.name} {state}" </h3>'),
                Math(self._latex_pdf), HTML(self._html_params()))

    def mean_E(self, R0, dr=0.01):
        """Mean E for each state from integration of the distance distribution
        """
        R_axis = self.get_r_axis(dr)
        E_from_R = fret.E_from_dist(R_axis, R0)
        if self.num_states == 1:
            R_pdf = self.pdf(R_axis, **self.params)
            E_mean_integral = np.trapz(E_from_R * R_pdf, R_axis)
        else:
            E_mean_integral = np.zeros(self.num_states)
            for i in range(self.num_states):
                p = {k: v[i] for k, v in self.params.items()}
                R_pdf = self.pdf(R_axis, **p)
                E_mean_integral[i] = np.trapz(E_from_R * R_pdf, R_axis)
        return E_mean_integral


class Gaussian(BaseDistribution):
    param_names = ('R_mean', 'R_sigma')
    _latex_pdf = (r'P(r) = \frac{1}{\sqrt{2 \pi \sigma^2}}'
                  r'\exp\left(-\frac{(r - \mu)^2}{2\sigma^2}\right) \\'
                  r'P(r) = 0 \quad\forall \; r \le 0; \qquad \mu > 3 \sigma > 0')

    @staticmethod
    def pdf(r, R_mean, R_sigma):
        u = (r - R_mean) / R_sigma
        area = R_sigma * np.sqrt(2 * np.pi)
        return np.exp(-(u**2) / 2) / area

    def r_max(self, n=5):
        return np.max(self.params['R_mean']) + n * np.max(self.params['R_sigma'])

    def _html_params(self):
        p = self.params
        return (f'<i>μ</i> = <code>R_mean</code> = {p["R_mean"]} <br>'
                f'<i>σ</i> = <code>R_sigma</code> = {p["R_sigma"]}')


class RadialGaussian(BaseDistribution):
    param_names = ('sigma', 'offset', 'mu')
    _latex_pdf = (r'P(r) = a^{-1}\; (r - r_0)^2 \;'
                  r'\exp\left[-\frac{[(r - r_0)- \mu]^2}{2\sigma^2}\right] \\'
                  r'a = \sqrt{\frac{\pi}{2}}\sigma(\sigma^2 + \mu^2) '
                  r'    + \mu \sigma^2 exp\left(-\frac{\mu^2}{2\sigma^2} \right)'
                  r'\mu \ge 0 \\'
                  r'P(r) = 0 \quad\forall \; r \le r_0\\')

    @staticmethod
    def pdf(r, sigma, offset, mu):
        sigma_sq = sigma * sigma
        mu_sq = mu * mu
        area = (np.sqrt(np.pi / 2) * sigma * (sigma_sq + mu_sq)
                * (1 - erf(-mu / (np.sqrt(2) * sigma)))
                + mu * sigma_sq * np.exp(- 0.5 * mu_sq / sigma_sq))
        r0 = (r - offset)
        valid = r0 > 0
        res = np.zeros_like(r)
        res[~valid] = 0
        r0_squared = r0[valid]**2
        r0_minus_mu_squared = (r0[valid] - mu)**2
        res[valid] = r0_squared * np.exp(-r0_minus_mu_squared / (2 * sigma_sq)) / area
        return res

    def _mean_peaks(self):
        mu, sigma = self.params['mu'], self.params['sigma']
        offset = self.params['offset']
        return (offset
                + np.sqrt(np.pi / 2) * mu * sigma * (mu**2 + 3 * sigma**2)
                * (1 - erf(-mu / (np.sqrt(2) * sigma)))
                + sigma**2 * (mu**2 + 2 * sigma**2) * np.exp(-0.5 * (mu / sigma)**2))

    def _std_dev_peaks(self):
        mu, sigma = self.params['mu'], self.params['sigma']
        M2 = (np.sqrt(np.pi / 2) * sigma * (mu**4 + 6 * mu**2 * sigma**2 + 3 * sigma**4)
              * (1 - erf(-mu / (np.sqrt(2) * sigma)))
              + sigma**2 * (mu**3 + 5 * mu * sigma**2))
        return M2 - self._mean_peaks()**2

    def r_max(self, n=5):
        return 20  # np.max(self._mean_peaks()) + n * np.max(self._std_dev_peaks())

    def _html_params(self):
        p = self.params
        return (f'<i>μ</i> = <code>mu</code> = {p["mu"]} <br>'
                f'<i>r₀</i> = <code>offset</code> = {p["offset"]} <br>'
                f'<i>σ</i> = <code>sigma</code> = {p["sigma"]}')


class GaussianChain(BaseDistribution):
    param_names = ('sigma', 'offset')
    _latex_pdf = (r'P(r) = a^{-1}\; (r - r_0)^2 \;'
                  r'\exp\left[-\frac{(r - r_0)^2}{2\sigma^2}\right] \\'
                  r'a = \sqrt{\frac{\pi}{2}}\sigma^3 \\'
                  r'P(r) = 0 \quad\forall \; r \le r_0\\')

    @staticmethod
    def pdf(r, sigma, offset):
        return RadialGaussian.pdf(r, sigma, offset, mu=0)

    def _mean_peaks(self):
        sigma, offset = self.params['sigma'], self.params['offset']
        return offset + 2 * sigma * np.sqrt(2 / np.pi)

    def _std_dev_peaks(self):
        sigma = self.params['sigma']
        return sigma * np.sqrt((3 * np.pi - 8) / np.pi)

    def r_max(self, n=5):
        return np.max(self._mean_peaks()) + n * np.max(self._std_dev_peaks())

    def _html_params(self):
        p = self.params
        return (f'<i>r₀</i> = <code>offset</code> = {p["offset"]} <br>'
                f'<i>σ</i> = <code>sigma</code> = {p["sigma"]}')


class WormLikeChain(BaseDistribution):
    param_names = ('L', 'lp', 'offset')
    _latex_pdf = (r'P(r) = c\,r^2 \, '
                  r'\left[ 1 - \left(\frac{r}{L}\right)^2 \right]^{-\frac{9}{2}} \,'
                  r'\exp\left[-\frac{9}{8}\frac{L}{l_p}'
                  r'\left( 1 - \left(\frac{r}{L}\right)^2 \right)^{-1} \right] \\'
                  r'P(r) = 0 \quad\forall\; r \le 0, \; r \ge L\\')

    @staticmethod
    def pdf(r, L, lp, offset):
        res = np.zeros_like(r)
        ro = r - offset
        valid = (ro < L) * (ro > 0)
        res[~valid] = 0
        ro_valid = ro[valid]
        F = 1 - (ro_valid / L)**2
        res[valid] = np.exp(- (9 / 8) * (L / lp) / F) * ro_valid**2 / F**(9 / 2)
        area = np.trapz(res, r)
        return res / area

    def r_max(self, n=5):
        return np.max(self.params['L']) + np.max(self.params['offset']) + n / 10

    def _html_params(self):
        p = self.params
        return (f'<i>L</i> = <code>L</code> = {p["L"]} <br>'
                f'<i>lp</i> = <code>lp</code> = {p["lp"]} <br>'
                f'<i>r₀</i> = <code>offset</code> = {p["offset"]}')


def _get_norm_cdf(du, u_max):
    temp = np.arange(0, u_max + du, du)
    u = np.hstack([-temp[::-1][:-1], temp])
    assert len(u) % 2 == 1, 'Size of `u` has to be an odd number.'
    idx_offset = (u.size - 1) // 2
    norm_pdf = Gaussian.pdf(u, R_mean=0, R_sigma=1)
    norm_cdf = np.cumsum(norm_pdf) * du
    return norm_cdf, idx_offset


def get_r_dist_distrib(du, u_max, dr, dd_params):
    """Computes R axis and index offset for mapping a distance distribution.

    Arguments:
        du (float): step-size for the x-axis of the unitary Gaussian PDF
        u_max (float): max range of x-axis for the unitary Gaussian PDF.
            The PFD is evaluated between -u_max and u_max.
        dr (float): step-size of the R axis of the new distance distribution.
            `dr` should smaller than `du`, for example `dr = 0.1 * du`.
        dd_params (dict): parameters of the new distance distribution.
            `dd_params['name']` must be a string with the name of the
            distance distribution (i.e. 'WLC' or 'radial_gaussian').
            The other elements need to match the parameter names of the
            chosen distance distribution.

    Returns:
        - r_dist (array): array of distances with same CDF values as
          a unitary Normal CDF evaluated on an array `u` going from
          `-u_max` to `u_max` with step `du`.
        - idx_offset (int): size of the positive side of the `u` array,
          not including 0.

    Note:
        Given a unitary (sigma = 1) normal random number `x`,
        a random number `R` from an arbitrary distribution can
        be obtained as::

            ix = int(round(x / du)) + idx_offset
            R = r_dist[ix]

        where `r_dist` and `idx_offset` are returned by this function,
        and `du` is the input argument to this function.
    """
    # CDF of the standard normal distribution with sigma = 1
    norm_cdf, idx_offset = _get_norm_cdf(du, u_max)
    # PDF and CDF of the new distance distribution
    r, dd_pdf = distribution(dd_params).get_pdf(dr, n=5)
    dd_cdf = np.cumsum(dd_pdf) * dr
    assert dd_cdf[-1] <= 1 + 1e-10, 'CDF is larger than 1!'
    assert 1 - dd_cdf[-1] < 1e-3, 'CDF axis range too small. '
    # Build the R-axis at positions matching the normal CDF values
    r_dist = np.interp(norm_cdf, dd_cdf, r)
    return r_dist, idx_offset



from .depi import (recolor_burstsph, mem, recolor_burstsph_cache,
                   save_params, load_params, validate_params)
from . import bva
from . import ctmc
from . import dist_distrib
from . import fret
from . import loader
from . import plotter

from ._version import get_versions
__version__ = get_versions()['version']
del get_versions



from functools import partial
from pathlib import Path
import numpy as np
import pandas as pd
import json
from joblib import Memory

from . import dist_distrib as dd
from . import fret
from . import depi_cy


mem = Memory(cachedir='joblib_cache')


@mem.cache
def recolor_burstsph_cache(timestamp, seed=1, **params):
    """Cached version of :func:`recolor_burstsph`."""
    from randomgen import RandomGenerator, Xoroshiro128
    rg = RandomGenerator(Xoroshiro128(seed))
    burstsph = recolor_burstsph(timestamp, rg=rg, **params)
    return burstsph


def save_params(fname, params, bounds=None, fit_name=None):
    """Save the simulation parameters `params` to disk."""
    if not fname.lower().endswith('.json'):
        fname = fname + '.json'
    if Path(fname).exists():
        raise IOError(f'File {fname} aready exists. '
                      f'Please move it before saving with the same name.')
    if bounds is not None:
        params = {**params, 'bounds': bounds}
    if fit_name is not None:
        params = {**params, 'fit_name': fit_name}
    with open(fname, 'wt') as f:
        json.dump(params, f, ensure_ascii=False, indent=4)


def load_params(fname):
    """Load the simulation parameters from file `fname`."""
    if not fname.lower().endswith('.json'):
        fname = fname + '.json'
    with open(fname) as f:
        params = json.load(f)
    bounds = params.pop('bounds', None)
    fit_name = params.pop('fit_name', None)
    return params, bounds, fit_name


def validate_params(params):
    """Validate consistency of simulation parameters."""
    d = dd.distribution(params)  # validates the distance-distribution parameters
    _check_args(params['τ_relax'], params['ndt'], params['α'], d.num_states)
    _get_multi_lifetime_components(params['τ_D'], params.get('D_fract'), 'D')
    _get_multi_lifetime_components(params['τ_A'], params.get('A_fract'), 'A')


def _check_args(τ_relax, ndt, α, num_states):
    """Check input arguments for recoloring simulation."""
    if all(np.atleast_1d(np.asarray(τ_relax)) == 0) and ndt > 0:
        raise ValueError('When τ_relax = 0 also ndt needs to be 0 '
                         'in order to avoid a 0 time-step size.')
    if α <= 0:
        raise ValueError(f'α needs to be strictly positive. It is {α}.')
    if np.size(τ_relax) != num_states:
        msg = (f'Parameter τ_relax ({τ_relax}) needs to have same size as '
               f'the number of states ({num_states})')
        raise TypeError(msg)


def _get_multi_lifetime_components(τ_X, X_fract, label='D'):
    """Check and type-convert params for D or A fluorescence lifetime components.
    """
    k_X = 1 / np.atleast_1d(τ_X)
    num_X_states = k_X.size
    if num_X_states > 1 and X_fract is None:
        msg = (f'When there is more than one {label} lifetime component, '
               'you need to specify the parameter `D_fract`.')
        raise TypeError(msg)
    X_fract = np.atleast_1d(X_fract).astype('float')
    if X_fract.size != num_X_states:
        msg = (f'Arrays `{label}_fract` ({X_fract}) and τ_{label} '
               f'({τ_X}) need to have the same size.')
        raise ValueError(msg)
    if not np.allclose(X_fract.sum(), 1, rtol=0, atol=1e-10):
        msg = (f'{label}_fract must sum to 1. It sums to {X_fract.sum()} instead.')
        raise ValueError(msg)
    return k_X, X_fract


def recolor_burstsph(
        timestamps, *, R0, τ_relax, δt, τ_D, τ_A, D_fract=1., A_fract=1.,
        bg_rate_d=0, bg_rate_a=0, ts_unit=1e-9, tcspc_range=50,
        prob_A_dark=0, τ_A_dark=0, gamma=1.0, lk=0., dir_ex_t=0.,
        rg=None, chunk_size=1000, α=0.05, ndt=10,
        **dd_model):
    """Recolor burst photons with Ornstein–Uhlenbeck D-A distance diffusion.

    Simulate D-A distances diffusing with a Gaussian or non-Gaussian distance
    distribution using an Ornstein–Uhlenbeck (OU) process with relaxation
    time `τ_relax`.
    Each input timestamp is the D excitation time. The D de-excitation
    happens either via D radiative emission (`τ_D`) or via FRET to A
    (distance dependent). The parameter `δt` is the time-step of the
    D de-excitation simulation which takes into account the D-A diffusion.
    When `τ_relax` is comparable or smaller than `τ_D`, the resulting FRET
    histogram is biased toward higher FRET values, a phenomenon known
    as diffusion-enhanced FRET (Beechem-Haas BJ1989).

    The input dictionary `dd_model` sets the distance distribution model.
    This dictionary have a key "name" which should be one of the strings:
    'gaussian', 'wlc', 'gaussian_chain', 'radial_gaussian'.
    Multi-state models can be specified when the distance distribution
    parameters are arrays instead of scalars. In this case also `τ_relax`
    needs to be an array (of same length).

    The D and A fluorescence decays can be single of multi-exponential.
    When `τ_D` or `τ_A` are scalars the decays are mono-exponential.
    When `τ_D` or `τ_A` are arrays the argument `D_fract` and `A_fract`
    specify the fractions of the different exponential components.

    Arguments:
        timestamps (pandas.Series): macrotimes of photons to be recolored.
            The index needs to have two levels: ('burst', 'ph').
        R0 (float): Förster radious of the D-A pair
        τ_relax (float): relaxation time of the OU process
        δt (float): nanotime (TCSPC) resolution, same units as
            `timestamps`. The actual time-bin `dt` may be smaller than `δt`
            because of a dependence on `τ_relax` (see argument `ndt`).
        α (float): sets the threshold above which switching
            from approximated probability to a CDF one.
        ndt (float): sets the max allowed time-step `δt`, potentially overriding
            the user supplied `δt`. If `τ_relax < ndt * δt`,
            then `δt` is reduced to `τ_relax / ndt`. The
            `δt` adjustment depends only on the input argument `τ_relax`
            and is performed only once at the beginning of the simulation.
            To avoid any `δt` adjustment, use `ndt = 0`.
        rg (None or RandomGenerator): random number generator object,
            usually `numpy.random.RandomState()`. If None, use
            `numpy.random.RandomState()` with no seed. Use this to pass
            an RNG initialized with a specific seed or to choose a
            RNG other than numpy's default Mersen Twister MT19937.
        chunk_size (int): request random numbers in chunks of this size

    Returns:
        burstsph (pandas.DataFrame): recolored photon data. Columns include: 'timestamp'
            (same as input timestamps); 'nanotime' (simulated TCSPC nanotimes);
            'stream' (color or the photon); 'R_ph' the D-A distance at
            de-excitation time; 'state' the state for each photon.
    """
    name = dd_model['name'].lower()
    dd.assert_valid_model_name(name)
    if name.startswith('gauss'):
        # Gaussian distance distributions
        func = recolor_burstsph_OU_gauss_R
    else:
        # Any non-Gaussian distance distribution
        func = recolor_burstsph_OU_dist_distrib
    return func(
        timestamps, R0=R0, gamma=gamma, lk=lk, dir_ex_t=dir_ex_t, τ_relax=τ_relax, δt=δt,
        τ_D=τ_D, D_fract=D_fract, τ_A=τ_A, A_fract=A_fract,
        prob_A_dark=prob_A_dark, τ_A_dark=τ_A_dark,
        bg_rate_d=bg_rate_d, bg_rate_a=bg_rate_a, ts_unit=ts_unit, tcspc_range=tcspc_range,
        rg=rg, chunk_size=chunk_size, α=α, ndt=ndt, dd_params=dd_model)


def recolor_burstsph_OU_dist_distrib(
        timestamps, *, R0, gamma, lk, dir_ex_t, τ_relax, dd_params, δt,
        τ_D, τ_A, D_fract=1., A_fract=1., bg_rate_d, bg_rate_a, ts_unit, tcspc_range,
        prob_A_dark=0, τ_A_dark=0,
        rg=None, chunk_size=1000, α=0.05, ndt=10):
    """Recoloring simulation with non-Gaussian distance distribution.
    """
    if rg is None:
        rg = np.random.RandomState()
    d = dd.distribution(dd_params)  # validates the distance-distribution parameters
    _check_args(τ_relax, ndt, α, d.num_states)
    k_D, D_fract = _get_multi_lifetime_components(τ_D, D_fract, 'D')
    k_A, A_fract = _get_multi_lifetime_components(τ_A, A_fract, 'A')
    ts = timestamps.values
    # Extract the fixed parameters that do not depend on the number of states
    dd_params = dd_params.copy()
    du = dd_params.pop('du')
    u_max = dd_params.pop('u_max')
    dr = dd_params.pop('dr')
    if d.num_states == 1:
        S_ph = None
        p = {k: v if np.isscalar(v) else v[0]
             for k, v in {'τ_relax': τ_relax, **dd_params}.items()}
        # Compute R-axis for the 1-state distance distribution
        r_dd, idx_offset_dd = dd.get_r_dist_distrib(
            du=du, u_max=u_max, dr=dr, dd_params=dd_params)
        # Run the 1-state recoloring simulation
        R_ph, T_ph = depi_cy.sim_DA_from_timestamps2_p2_dist_cy(
            ts, δt, k_D, D_fract, R0, p['τ_relax'], r_dd, idx_offset_dd, du,
            rg=rg, chunk_size=chunk_size, alpha=α, ndt=ndt)
    else:
        funcs = {1: depi_cy.sim_DA_from_timestamps2_p2_2states_dist_cy,
                 2: depi_cy.sim_DA_from_timestamps2_p2_Nstates_dist_cy}
        # Compute R-axis for the multi-state distance distribution
        r_dd_list, idx_offset_list = [], []
        for i in range(d.num_states):
            dd_params_i = {k: v[i] for k, v in dd_params.items() if k != 'name'}
            dd_params_i['name'] = dd_params['name']
            r_dd, idx_offset = dd.get_r_dist_distrib(
                du=du, u_max=u_max, dr=dr, dd_params=dd_params_i)
            r_dd_list.append(r_dd)
            idx_offset_list.append(idx_offset)
        r_dd = np.vstack(r_dd_list)
        idx_offset_dd = np.array(idx_offset_list, dtype='int64')
        # Run the multi-state recoloring simulation
        R_ph, T_ph, S_ph = funcs[d.k_s.ndim](
            ts, δt, k_D, D_fract, R0, np.asfarray(τ_relax), d.k_s, r_dd, idx_offset_dd, du,
            rg=rg, chunk_size=chunk_size, alpha=α, ndt=ndt)
    burstsph_all = _make_burstsph_df(timestamps, T_ph, R_ph, S_ph)
    burstsph, bg = _select_background(burstsph_all, bg_rate_d=bg_rate_d, bg_rate_a=bg_rate_a,
                                      ts_unit=ts_unit, tcspc_range=tcspc_range, rg=rg)
    burstsph = _color_photons(burstsph, R0, gamma=gamma, lk=lk, dir_ex_t=dir_ex_t, rg=rg)
    burstsph = _recolor_A_blinking(burstsph, prob_A_dark, τ_A_dark, lk, τ_D, D_fract, rg)
    burstsph = _add_acceptor_nanotime(burstsph, τ_A, A_fract, rg)
    burstsph = _merge_ph_and_bg(burstsph_all, burstsph, bg)
    return burstsph


def recolor_burstsph_OU_gauss_R(
        timestamps, *, R0, gamma, lk, dir_ex_t, τ_relax, dd_params, δt,
        τ_D, τ_A, D_fract=1., A_fract=1., bg_rate_d, bg_rate_a, ts_unit, tcspc_range,
        prob_A_dark=0, τ_A_dark=0,
        rg=None, chunk_size=1000, α=0.05, ndt=10, cdf=True):
    """Recoloring simulation with Gaussian distance distribution.
    """
    if rg is None:
        rg = np.random.RandomState()
    d = dd.distribution(dd_params)  # validates the distance-distribution parameters
    _check_args(τ_relax, ndt, α, d.num_states)
    k_D, D_fract = _get_multi_lifetime_components(τ_D, D_fract, 'D')
    k_A, A_fract = _get_multi_lifetime_components(τ_A, A_fract, 'A')
    ts = timestamps.values
    if d.num_states == 1:
        S_ph = None
        if cdf:
            func = depi_cy.sim_DA_from_timestamps2_p2_cy
        else:
            func = depi_cy.sim_DA_from_timestamps2_p_cy
        R_ph, T_ph = func(
            ts, δt, k_D, D_fract, R0, dd_params['R_mean'], dd_params['R_sigma'],
            τ_relax, rg=rg, chunk_size=chunk_size, alpha=α, ndt=ndt)
    else:
        funcs = {1: depi_cy.sim_DA_from_timestamps2_p2_2states_cy,
                 2: depi_cy.sim_DA_from_timestamps2_p2_Nstates_cy}
        params = (np.asfarray(dd_params['R_mean']),
                  np.asfarray(dd_params['R_sigma']),
                  np.asfarray(τ_relax), np.asfarray(d.k_s))
        R_ph, T_ph, S_ph = funcs[d.k_s.ndim](
            ts, δt, k_D, D_fract, R0, *params,
            rg=rg, chunk_size=chunk_size, alpha=α, ndt=ndt)
    burstsph_all = _make_burstsph_df(timestamps, T_ph, R_ph, S_ph)
    burstsph, bg = _select_background(burstsph_all, bg_rate_d=bg_rate_d, bg_rate_a=bg_rate_a,
                                      ts_unit=ts_unit, tcspc_range=tcspc_range, rg=rg)
    burstsph = _color_photons(burstsph, R0, gamma=gamma, lk=lk, dir_ex_t=dir_ex_t, rg=rg)
    burstsph = _recolor_A_blinking(burstsph, prob_A_dark, τ_A_dark, lk, τ_D, D_fract, rg)
    burstsph = _add_acceptor_nanotime(burstsph, τ_A, A_fract, rg)
    burstsph = _merge_ph_and_bg(burstsph_all, burstsph, bg)
    return burstsph


def _recolor_A_blinking(burstsph, prob_A_dark, τ_A_dark, lk, τ_D, D_fract, rg):
    if prob_A_dark == 0:
        burstsph['A_dark_ph'] = False
        return burstsph
    # Acceptor-dye emitted photons (no D leakage), before A photo-blinking
    Aem = (burstsph.stream == 'DexAem') & ~burstsph.leak_ph
    num_Aem = Aem.sum()
    # Simulate acceptor bright to dark transitions
    u = rg.rand(num_Aem)
    A_dark = u <= prob_A_dark
    # Recolor and label photons emitted during A dark state
    # Generate a tau for each `Aem` photon, then discard it unless the photon is "dark"
    A_dark_lifetimes = rg.exponential(scale=τ_A_dark, size=num_Aem)
    ts = burstsph.loc[Aem, 'timestamp'].values
    A_dark = _generate_acceptor_dark_state(A_dark, ts, A_dark_lifetimes)
    num_A_dark = A_dark.sum()
    assert A_dark.size == num_Aem
    burstsph['A_dark_ph'] = False
    burstsph.loc[Aem, 'A_dark_ph'] = A_dark
    burstsph.loc[burstsph.A_dark_ph, 'A_ch'] = False
    assert num_A_dark == burstsph.A_dark_ph.sum()
    assert burstsph.loc[burstsph.A_dark_ph, 'A_dark_ph'].all()
    assert not burstsph.loc[burstsph.A_dark_ph, 'A_ch'].any()
    # Set the D fluorescence lifetime for blinked photons
    burstsph.loc[burstsph.A_dark_ph, 'nanotime'] = _calc_intrisic_nanotime(
        num_A_dark, τ_D, D_fract, rg)
    # Simulate residual D leakage
    if lk > 0:
        # NOTE: leaked photons go into the A channel but retain D nanotime
        new_lk_ph = rg.binomial(1, p=lk, size=num_A_dark)
        burstsph.loc[burstsph.A_dark_ph, 'leak_ph'] |= new_lk_ph
        burstsph.loc[burstsph.leak_ph, 'A_ch'] = True
    # Recompute the categorical "stream" column from A_ch
    burstsph = _set_ph_stream_column(burstsph, burstsph.A_ch)
    return burstsph


def _generate_acceptor_dark_state(A_dark, ts, A_dark_times):
    A_dark = memoryview(A_dark)
    ts = memoryview(ts)
    A_dark_times = memoryview(A_dark_times)
    dark = False
    for i in range(len(A_dark)):
        if A_dark[i]:
            # switch A to the dark state
            dark = True
            dark_start = ts[i]
            dark_time = A_dark_times[i]
        if dark:
            A_dark[i] = True  # label photon as A dark-state
            dark_time_elapsed = ts[i] - dark_start
            if dark_time_elapsed >= dark_time:
                # go back to the bright state
                dark = False
    return np.asarray(A_dark)


def _make_burstsph_df(timestamps, T_ph, R_ph, S_ph=None):
    burstsph = pd.DataFrame(timestamps)
    burstsph['nanotime'] = T_ph
    burstsph['R_ph'] = R_ph
    if S_ph is not None:
        burstsph['state'] = S_ph
    return burstsph


def _select_background(df, bg_rate_d, bg_rate_a, ts_unit, tcspc_range, rg):
    if bg_rate_d == 0 and bg_rate_a == 0:
        return df, pd.DataFrame(columns=['bg_a', 'bg_d'], dtype=bool)
    func = partial(_background_per_burst, bg_rate_d=bg_rate_d, bg_rate_a=bg_rate_a,
                   ts_unit=ts_unit, rg=rg)
    bg = df.timestamp.groupby('burst').apply(func)
    bg_mask = bg.bg_a | bg.bg_d
    bg['nanotime'] = np.nan
    bg.loc[bg_mask, 'nanotime'] = rg.rand(bg_mask.sum()) * tcspc_range
    return df.loc[~bg_mask].copy(), bg


def _background_per_burst(burst, bg_rate_d, bg_rate_a, ts_unit, rg):
    burst = burst.reset_index('burst', drop=True)
    bsize = burst.shape[0]
    bwidth = (burst.iloc[-1] - burst.iloc[0]) * ts_unit
    bg_burst = (bg_rate_d + bg_rate_a) * bwidth
    assert bg_burst < bsize
    prob_A_ph_bg = bg_rate_a / (bg_rate_a + bg_rate_d)
    prob_bg_ph = bg_burst / bsize
    bg_ph = rg.binomial(1, p=prob_bg_ph, size=bsize)
    bg_ph_index = np.where(bg_ph)[0]
    a_bg_ph = rg.binomial(1, p=prob_A_ph_bg, size=bg_ph_index.size).astype(bool)
    a_bg_ph_index = bg_ph_index[a_bg_ph]
    d_bg_ph_index = list(set(bg_ph_index) - set(a_bg_ph_index))
    bg = pd.DataFrame(index=burst.index, columns=['bg_a', 'bg_d'], dtype=bool)
    bg.loc[:] = False
    bg.iloc[a_bg_ph_index, 0] = True  # assign column 'bg_a'
    bg.iloc[d_bg_ph_index, 1] = True  # assign column 'bg_d'
    return bg


def _merge_ph_and_bg(burstsph_all, burstsph, bg):
    if burstsph_all.shape[0] == burstsph.shape[0]:
        burstsph['bg_ph'] = False
        return burstsph
    # This way of merging bg into a bigger DataFrame is ugly and brittle
    # there must be a better way!
    bg_mask = bg.bg_a | bg.bg_d
    bool_cols = ['A_ch', 'leak_ph', 'dir_ex_ph', 'A_dark_ph']
    for col in bool_cols:
        burstsph_all[col] = False  # create the column
    burstsph_all['bg_ph'] = False  # create the bg_ph column
    burstsph_all.loc[bg_mask, 'bg_ph'] = True
    for col in ['nanotime'] + bool_cols:
        burstsph_all.loc[~bg_mask, col] = burstsph[col]
    burstsph_all.loc[bg_mask, 'nanotime'] = bg.loc[bg_mask, 'nanotime']
    burstsph_all.loc[bg_mask, 'A_ch'] = bg.loc[bg_mask, 'bg_a']
    burstsph_all = _set_ph_stream_column(burstsph_all, burstsph_all.A_ch)
    return burstsph_all


def _set_ph_stream_column(df, A_ch):
    df['stream'] = pd.Categorical.from_codes(A_ch, categories=["DexDem", "DexAem"])
    return df


def _color_photons(df, R0, gamma, lk, dir_ex_t, rg):
    # Color photons (either D or A)
    num_ph = df.shape[0]
    E = fret.E_from_dist(df.R_ph, R0)
    Eraw = fret.uncorrect_E_gamma_leak_dir(E, gamma=gamma, leakage=lk, dir_ex_t=dir_ex_t)
    A_em = rg.binomial(1, p=Eraw, size=num_ph).astype(bool)
    D_em = ~A_em
    NA = A_em.sum()
    ND = A_em.size - NA
    assert D_em.sum() == ND
    df['A_ch'] = A_em
    df = _set_ph_stream_column(df, A_em)
    # These values are shot-noise free
    ND_theor = num_ph * (1 - Eraw)
    NA_theor = num_ph * Eraw

    # Assign D leakage photons
    # NOTE: fract_lk_ph is equal to lk*ND / NA but without the shot noise
    fract_lk_ph = lk * ND_theor[A_em] / NA_theor[A_em]
    assert fract_lk_ph.size == NA
    leaked_photons = rg.binomial(1, p=fract_lk_ph, size=NA).astype(bool)
    df['leak_ph'] = False
    df.loc[A_em, 'leak_ph'] = leaked_photons
    A_em_nolk = A_em & (~df.leak_ph)
    assert df.leak_ph.sum() == leaked_photons.sum(), f'{df.leak_ph.sum()}, {leaked_photons.sum()}'

    # Assign A direct excitation photons
    Lk = leaked_photons.sum()
    nd = ND_theor[A_em_nolk]
    na_raw = NA_theor[A_em_nolk]
    na = (na_raw - dir_ex_t * gamma * nd - Lk) / (1 + dir_ex_t)
    fract_dir_ex_ph = dir_ex_t * (gamma * nd + na) / (na_raw - Lk)
    assert fract_dir_ex_ph.size == NA - leaked_photons.sum()
    dir_ex_photons = rg.binomial(1, p=fract_dir_ex_ph,
                                 size=(NA - leaked_photons.sum())).astype(bool)
    num_dir_ex_ph = dir_ex_photons.sum()
    df['dir_ex_ph'] = False
    df.loc[A_em_nolk, 'dir_ex_ph'] = dir_ex_photons
    assert df.dir_ex_ph.sum() == num_dir_ex_ph, f'{df.dir_ex_ph.sum()}, {num_dir_ex_ph}'
    assert A_em[df.leak_ph].all()
    assert A_em[df.dir_ex_ph].all()
    return df


def _add_acceptor_nanotime(df, τ_A, A_fract, rg):
    """Modify the D nanotime by adding the A components.
    Leaked photons are not touched. This function sets the nanotime
    of A-direct-excitation photons to 0 before adding the
    A fluorescence components.
    """
    # Save original D-deexcitation nanotimes
    df['nanotime_d'] = df['nanotime'].copy()
    # Set the initial nanotime of A direct excitation photons to 0
    nanot_dir_ex = df.nanotime.copy()
    nanot_dir_ex[df.dir_ex_ph] = 0
    df.nanotime = nanot_dir_ex
    # Add A lifetimes to non-leaked photons
    A_mask = df.A_ch & ~df.leak_ph
    nanotime = _calc_intrisic_nanotime(A_mask.sum(), τ_A, A_fract, rg)
    df.loc[A_mask, 'nanotime'] += nanotime
    return df


def _calc_intrisic_nanotime(num_ph, τ_X, X_fract, rg):
    """Generate the nanotime for single or multi-componet decays.
    """
    if np.size(τ_X) == 1:
        nanotime = rg.exponential(scale=τ_X, size=num_ph)
    else:
        num_X_comps = len(τ_X)
        components = rg.choice(num_X_comps, size=num_ph, p=X_fract)
        nanotime = np.zeros(num_ph)
        for i, τ_X_i in enumerate(τ_X):
            # Compute nanotimes for component `i` of the fluorescence decay
            component_i = components == i
            nanotime[component_i] = rg.exponential(scale=τ_X_i, size=component_i.sum())
    return nanotime



import numpy as np
import scipy.linalg as la


def G_from_Q(Q):
    G = Q.copy()
    G[np.diag_indices_from(G)] = -leaving_rates(Q)
    return G


def Q_from_G(G):
    Q = G.copy()
    Q[np.diag_indices_from(Q)] = 0
    return Q


def leaving_rates(G):
    Q = Q_from_G(G)
    return Q.sum(1)


def entering_rates(G):
    Q = Q_from_G(G)
    return Q.sum(0)


def P_from_G(G):
    Q = Q_from_G(G)
    return Q / leaving_rates(Q)[np.newaxis, :].T


def CT_stationary_distribution(G):
    """
    Continuous-time stationary distribution computed from
    the left eigenvector of the zero eigen value.
    """
    λ, V = la.eig(G.T)
    i_zero_left_eig = np.where(np.abs(λ) < 1e-12)[0]
    π = (V[:, i_zero_left_eig].T[0] / V[:, i_zero_left_eig].sum()).real
    return π


def CT_stationary_distribution_from_psi(G, ψ):
    """
    Continuous-time stationary distribution computed from
    the stationary distribution ψ of the embedded chain.
    """
    v_inv = 1 / leaving_rates(G)
    π = (ψ * v_inv) / np.dot(ψ, v_inv)
    return π


def DT_embedded_stationary_distribution_from_pi(G, π):
    """
    Discrete-time stationary distribution of the embedded chain
    computed from the continous-time stationary distribution π.
    """
    v = leaving_rates(G)
    ψ = π * v / np.dot(π, v)
    return ψ


def DT_embedded_stationary_distribution(G, n=50):
    """
    Discrete-time stationary distribution of the embedded chain
    computed by iterative transition matrix multiplication.
    """
    P = P_from_G(G)
    ψ0 = np.ones((1, P.shape[0]))
    ψ = ψ0 @ np.linalg.matrix_power(P, n)
    ψ /= ψ.sum()
    return ψ


def transition_matrix_t(t, G):
    λ, V = la.eig(G)
    P_t = (V @ np.diag(np.exp(λ * t)) @ la.inv(V)).real
    return P_t


def decompose(G):
    λ, V = la.eig(G)
    V_inv = la.inv(V)
    return λ, V, V_inv


def occupancy(t, α, G, λ=None, V=None, V_inv=None):
    if λ is None:
        λ, V = la.eig(G)
    if V_inv is None:
        V_inv = la.inv(V)
    P_t_matrix = (V @ np.diag(np.exp(λ * t)) @ V_inv).real
    P_t = α @ P_t_matrix
    return P_t


def occupancy_vs_t(t, α, G, λ=None, V=None, V_inv=None):
    if λ is None:
        λ, V = la.eig(G)
    if V_inv is None:
        V_inv = la.inv(V)
    t = np.asarray(t)
    P_t = np.zeros((t.size, G.shape[0]))
    for i, ti in enumerate(t):
        P_ti = (V @ np.diag(np.exp(λ * ti)) @ V_inv).real
        P_t[i] = α @ P_ti
    return P_t



from pathlib import Path
import json
import numpy as np
import pandas as pd
from pandas.api.types import CategoricalDtype

stream_map = {0: 'DexDem', 1: 'DexAem', 2: 'AexDem', 3: 'AexAem'}
stream_dtype = CategoricalDtype(categories=stream_map.values())


def fname_from_pattern(pattern, folder):
    matching_files = [f for f in Path(folder).glob(pattern)]
    if len(matching_files) == 0:
        msg = f'No file matching the pattern "{pattern}"'
        raise FileNotFoundError(msg)
    elif len(matching_files) > 1:
        msg = f'Multiple files matching the pattern "{pattern}":\n'
        for f in matching_files:
            msg += f'- {f}\n'
        raise FileNotFoundError(msg)
    else:
        return matching_files[0]


def load_burst_and_ph_data(name, folder, pop):
    suffixb = '_%s_bursts'
    suffixph = '_%s_burst_photons'
    if isinstance(name, str):
        patternph = f'{name}_merge{suffixph % pop}.csv'
    else:
        conc = name
        patternph = f'*{conc:.1f}M_merge{suffixph % pop}.csv'
    fname = fname_from_pattern(patternph, folder)
    burstsph = pd.read_csv(fname, skiprows=1, index_col=(0, 1))
    burstsph.stream = burstsph.stream.astype(stream_dtype)
    header = fname.read_text().split('\n')[0]
    meta = json.loads(header)
    meta['timestamp_unit_hw'] = meta['timestamp_unit']
    scale = meta['timestamp_unit'] * 1e9
    scale = int(scale) if round(scale) == scale else scale
    burstsph.timestamp *= scale
    meta['timestamp_unit'] = 1e-9
    print(f'- Loaded photon-data "{fname}"')

    if isinstance(name, str):
        patternb = f'{name}_merge{suffixb % pop}.csv'
    else:
        patternb = f'*{conc:.1f}M_merge{suffixb % pop}.csv'
    fname = fname_from_pattern(patternb, folder)
    print(fname)
    bursts = pd.read_csv(fname, index_col=0)
    istart = np.hstack([[0], np.cumsum(bursts.size_raw)[:-1]])
    istop = np.cumsum(bursts.size_raw.values)
    assert istart.size == istop.size
    bursts['istart'] = istart
    bursts['istop'] = istop
    print(f'- Loaded burst data "{fname}"')

    # Tests
    assert (np.diff(burstsph.timestamp) >= 0).all()
    num_bursts = np.unique(burstsph.reset_index('burst').burst).size
    assert bursts.shape[0] == num_bursts
    assert (burstsph.groupby('burst')['timestamp'].count() == bursts.size_raw).all()
    return bursts, burstsph, meta



from math import exp
import numpy as np
import pandas as pd
import numba

from . import depi_cy
from .depi_cy import ou_single_step_cy
from . import fret
from . import ctmc


#
# - FOLLOWING FUNCTIONS ARE SLOW PYTHON VERSIONS FOR TESTING ONLY - - - -
#

def sim_DA_from_timestamps2_p2(timestamps, dt, k_D, R0, R_mean, R_sigma,
                               tau_relax, rg, chunk_size=1000,
                               alpha=0.05, ndt=10):
    """
    Recoloring using fixed dt, emission uses exponential CDF.
    For efficiency, random numbers are computed in blocks.

    The assumption here is that since tau_relax >> dt, distance and FRET
    will not change during dt, so the transition probability can be computed
    from the exponential CDF.
    """
    if tau_relax < ndt * dt:
        dt = tau_relax / ndt
        print(f'WARNING: Reducing dt to {dt:g} '
              f'[tau_relax = {tau_relax}]')
    R = rg.randn() * R_sigma + R_mean
    t0 = 0
    nanotime = 0
    # Array flagging photons as A (1) or D (0) emitted
    A_ph = np.zeros(timestamps.size, dtype=bool)
    # Instantaneous D-A distance at D de-excitation time
    R_ph = np.zeros(timestamps.size, dtype=np.float64)
    # Time of D de-excitation relative to the last timestamp
    T_ph = np.zeros(timestamps.size, dtype=np.float64)
    iN = chunk_size - 1  # value to get the first chunk of random numbers
    for iph, t in enumerate(timestamps):
        # each cycle starts with a new photon timestamp `t`
        # excitation time is `t`, emission time is `t + nanotime`
        delta_t = t - t0
        if delta_t < 0:
            # avoid negative delta_t possible when when two photons have
            # the same macrotime
            delta_t = 0
            t = t0
        # Compute the D-A distance at the "excitation time"
        iN += 1
        if iN == chunk_size:
            Na = memoryview(rg.randn(chunk_size))
            Pa = memoryview(rg.rand(chunk_size))
            iN = 0
        N = Na[iN]
        p = Pa[iN]
        R = ou_single_step_cy(R, delta_t, N, R_mean, R_sigma, tau_relax)
        nanotime = 0
        # loop through D-A diffusion steps with a fixed time-step dt
        # until D de-excitation by photon emission or energy transfer to A
        while True:
            k_ET = k_D * (R0 / R)**6
            k_emission = k_ET + k_D
            d_prob_ph_em = dt * k_emission
            if d_prob_ph_em > alpha:
                d_prob_ph_em = 1 - exp(-d_prob_ph_em)
            if d_prob_ph_em >= p:
                break   # break out of the loop when the photon is emitted
            nanotime += dt
            iN += 1
            if iN == chunk_size:
                Na = memoryview(rg.randn(chunk_size))
                Pa = memoryview(rg.rand(chunk_size))
                iN = 0
            N = Na[iN]
            p = Pa[iN]
            # Update R following the OU process
            R = ou_single_step_cy(R, dt, N, R_mean, R_sigma, tau_relax)
        # photon emitted, let's decide if it is from D or A
        p_DA = p / d_prob_ph_em  # equivalent to rand(), but faster
        prob_A_em = k_ET / k_emission
        if prob_A_em >= p_DA:
            A_ph[iph] = True
        # time of D de-excitation by photon emission or energy transfer to A
        t0 = t + nanotime
        # save D-A distance at emission time
        R_ph[iph] = R
        # save time of emission relative to the excitation time `t`
        T_ph[iph] = nanotime
    return A_ph, R_ph, T_ph


def sim_DA_from_timestamps2_p2_2states(timestamps, dt_ref, k_D, R0, R_mean,
                                       R_sigma, tau_relax, k_s, rg,
                                       chunk_size=1000, alpha=0.05, ndt=10):
    """
    2-states recoloring using CDF in dt and with random number caching
    """
    dt = np.array([dt_ref] * 2, dtype=np.float64)
    for state in [0, 1]:
        if tau_relax[state] < ndt * dt[state]:
            dt[state] = tau_relax[state] / ndt
            print(f'WARNING: Reducing dt[{state}] to {dt[state]:g} '
                  f'[tau_relax[{state}] = {tau_relax[state]}]')
    # Array flagging photons as A (1) or D (0) emitted
    A_ph = np.zeros(timestamps.size, dtype=np.uint8)
    # Instantaneous D-A distance at D de-excitation time
    R_ph = np.zeros(timestamps.size, dtype=np.float64)
    # Time of D de-excitation relative to the last timestamp
    T_ph = np.zeros(timestamps.size, dtype=np.float64)
    # State for each photon
    S_ph = np.zeros(timestamps.size, dtype=np.uint8)
    peq = [k_s[1] / (k_s[0] + k_s[1]),
           k_s[0] / (k_s[0] + k_s[1])]
    k_s_sum = np.sum(k_s)
    t0 = 0
    nanotime = 0
    state = 0  # the two states are 0 and 1
    R = rg.randn() * R_sigma[state] + R_mean[state]
    iN = chunk_size - 1  # value to get the first chunk of random numbers
    for iph, t in enumerate(timestamps):
        # each cycle starts with a new photon timestamp `t`
        # excitation time is `t`, emission time is `t + nanotime`
        delta_t0 = t - t0
        delta_t = delta_t0 - nanotime
        if delta_t < 0:
            # avoid negative delta_t possible when when two photons have
            # the same macrotime
            delta_t = 0
            t = t0
        p_state = (1 - peq[state]) * np.exp(-(delta_t0 * k_s_sum)) + peq[state]
        u = rg.rand()
        #print(f'iph={iph}, state={state}, p_state={p_state}, u={u}, delta_t0={delta_t0}')
        # Inversion of u is for compatibility with N-state version
        if state == 1:
            u = 1 - u
        if p_state <= u:
            #print('   * state change')
            state = 0 if state == 1 else 1
            R = rg.randn() * R_sigma[state] + R_mean[state]
        # Compute the D-A distance at the "excitation time"
        iN += 1
        if iN == chunk_size:
            Na = memoryview(rg.randn(chunk_size))
            Pa = memoryview(rg.rand(chunk_size))
            iN = 0
        N = Na[iN]
        p = Pa[iN]
        R = ou_single_step_cy(R, delta_t, N, R_mean[state], R_sigma[state],
                              tau_relax[state])
        nanotime = 0
        # loop through D-A diffusion steps with a fixed time-step dt
        # until D de-excitation by photon emission or energy transfer to A
        while True:
            k_ET = k_D * (R0 / R)**6
            k_emission = k_ET + k_D
            d_prob_ph_em = k_emission * dt[state]  # prob. of emission in dt
            if d_prob_ph_em > alpha:
                d_prob_ph_em = 1 - exp(-d_prob_ph_em)
            if d_prob_ph_em >= p:
                break   # break out of the loop when the photon is emitted
            nanotime += dt[state]
            iN += 1
            if iN == chunk_size:
                Na = memoryview(rg.randn(chunk_size))
                Pa = memoryview(rg.rand(chunk_size))
                iN = 0
            N = Na[iN]
            p = Pa[iN]
            R = ou_single_step_cy(R, dt[state], N, R_mean[state], R_sigma[state],
                                  tau_relax[state])
        # photon emitted, let's decide if it is from D or A
        p_DA = p / d_prob_ph_em  # equivalent to rand(), but faster
        prob_A_em = k_ET / k_emission
        if prob_A_em >= p_DA:
            A_ph[iph] = 1
        # time of D de-excitation by photon emission or energy transfer to A
        t0 = t
        # save D-A distance at emission time
        R_ph[iph] = R
        # save time of emission relative to the excitation time `t`
        T_ph[iph] = nanotime
        # Save state for current photon
        S_ph[iph] = state
    return A_ph, R_ph, T_ph, S_ph


def sim_DA_from_timestamps2_p2_Nstates(timestamps, dt_ref, k_D, R0, R_mean,
                                       R_sigma, tau_relax, K_s, rg,
                                       chunk_size=1000, alpha=0.05, ndt=10):
    """
    2-states recoloring using CDF in dt and with random number caching
    """
    assert K_s.shape[0] == K_s.shape[1], 'K_s needs to be a square matrix.'
    num_states = K_s.shape[0]
    dt = np.array([dt_ref] * num_states, dtype=np.float64)
    for state in range(num_states):
        if tau_relax[state] < ndt * dt[state]:
            dt[state] = tau_relax[state] / ndt
            print(f'WARNING: Reducing dt[{state}] to {dt[state]:g} '
                  f'[tau_relax[{state}] = {tau_relax[state]}]')
    # Array flagging photons as A (1) or D (0) emitted
    A_ph = np.zeros(timestamps.size, dtype=np.uint8)
    # Instantaneous D-A distance at D de-excitation time
    R_ph = np.zeros(timestamps.size, dtype=np.float64)
    # Time of D de-excitation relative to the last timestamp
    T_ph = np.zeros(timestamps.size, dtype=np.float64)
    # State for each photon
    S_ph = np.zeros(timestamps.size, dtype=np.uint8)
    eigenval, V, V_inv = ctmc.decompose(K_s)
    t0 = 0
    nanotime = 0
    state = 0
    state_vector = np.zeros(num_states)
    state_vector[state] = 1
    R = rg.randn() * R_sigma[state] + R_mean[state]
    iN = chunk_size - 1  # value to get the first chunk of random numbers
    for iph, t in enumerate(timestamps):
        # each cycle starts with a new photon timestamp `t`
        # excitation time is `t`, emission time is `t + nanotime`
        delta_t0 = t - t0
        delta_t = delta_t0 - nanotime
        if delta_t < 0:
            # avoid negative delta_t possible when when two photons have
            # the same macrotime
            delta_t = 0
            t = t0
        p_states = ctmc.occupancy(
            delta_t0, state_vector, K_s, eigenval, V, V_inv)
        u = rg.rand()
        #print(f'iph={iph}, state={state}, p_states={p_states}, u={u}, delta_t0={delta_t0}')
        for s, p_state in enumerate(p_states):
            if u <= p_state:
                newstate = s
                break
            u -= p_state
        if newstate != state:
            #print('   * state change')
            state_vector[state] = 0
            state = newstate
            state_vector[state] = 1
            R = rg.randn() * R_sigma[state] + R_mean[state]
        # Compute the D-A distance at the "excitation time"
        iN += 1
        if iN == chunk_size:
            Na = memoryview(rg.randn(chunk_size))
            Pa = memoryview(rg.rand(chunk_size))
            iN = 0
        N = Na[iN]
        p = Pa[iN]
        R = ou_single_step_cy(R, delta_t, N, R_mean[state], R_sigma[state],
                              tau_relax[state])
        nanotime = 0
        # loop through D-A diffusion steps with a fixed time-step dt
        # until D de-excitation by photon emission or energy transfer to A
        while True:
            k_ET = k_D * (R0 / R)**6
            k_emission = k_ET + k_D
            d_prob_ph_em = k_emission * dt[state]  # prob. of emission in dt
            if d_prob_ph_em > alpha:
                d_prob_ph_em = 1 - exp(-d_prob_ph_em)
            if d_prob_ph_em >= p:
                break   # break out of the loop when the photon is emitted
            nanotime += dt[state]
            iN += 1
            if iN == chunk_size:
                Na = memoryview(rg.randn(chunk_size))
                Pa = memoryview(rg.rand(chunk_size))
                iN = 0
            N = Na[iN]
            p = Pa[iN]
            R = ou_single_step_cy(R, dt[state], N, R_mean[state], R_sigma[state],
                                  tau_relax[state])
        # photon emitted, let's decide if it is from D or A
        p_DA = p / d_prob_ph_em  # equivalent to rand(), but faster
        prob_A_em = k_ET / k_emission
        if prob_A_em >= p_DA:
            A_ph[iph] = 1
        # time of D de-excitation by photon emission or energy transfer to A
        t0 = t
        # save D-A distance at emission time
        R_ph[iph] = R
        # save time of emission relative to the excitation time `t`
        T_ph[iph] = nanotime
        # Save state for current photon
        S_ph[iph] = state
    return A_ph, R_ph, T_ph, S_ph


def recolor_burstsph_OU_gauss_R_approx(timestamps, R0, R_mean, R_sigma, τ_relax,
                                       τ_D, τ_A, δt, rg=None):
    """Recolor burst photons with Ornstein–Uhlenbeck D-A distance diffusion.

    This version is a more approximate version of
    :func:`recolor_burstsph_OU_gauss_R` using a fixed `δt`.

    Arguments:
        timestamps (pandas.Series): macrotimes of photons to be recolored.
            The index needs to have two levels: ('burst', 'ph').
        R0 (float): Förster radious of the D-A pair
        R_mean (float): mean D-A distance
        R_sigma (float): standard deviation of the distance distribution
        τ_relax (float): relaxation time of the OU process
        δt (float): time step for the diffusion-mediated D de-excitation.
            Same units as `timestamps`.
        rg (None or RandomGenerator): random number generator object,
            usually `numpy.random.RandomState()`. If None, use
            `numpy.random.RandomState()` with no seed. Use this to pass
            an RNG initialized with a specific seed or to choose a
            RNG other than numpy's default Mersen Twister MT19937.
    Returns:
        burstsph (pandas.DataFrame): DataFrame with 3 columns: 'timestamp'
            (same as input timestamps), 'nanotime' (simulated TCSPC nanotime)
            and 'stream' (color or the photon).
    """
    if rg is None:
        rg = np.random.RandomState()
    k_D = 1 / τ_D
    ts = timestamps.values
    A_em, R_ph, T_ph = depi_cy.sim_DA_from_timestamps2_cy(
        ts, δt, k_D, R0, R_mean, R_sigma, τ_relax, rg)
    A_mask = A_em.view(bool)
    T_ph = np.asarray(T_ph)
    T_ph[A_mask] += rg.exponential(scale=τ_A, size=A_mask.sum())
    burstsph_sim = pd.DataFrame(timestamps)
    burstsph_sim['nanotime'] = T_ph
    burstsph_sim['stream'] = (
        pd.Categorical.from_codes(A_em, categories=["DexDem", "DexAem"]))
    burstsph_sim['R_ph'] = R_ph
    return burstsph_sim


def sim_DA_from_timestamps(timestamps, δt, k_D, R0,
                           R_mean, R_sigma, τ_relax, rg):
    """
    Recoloring using a fixed δt and no random number caching
    """
    A_ph = np.zeros(timestamps.size, dtype=bool)
    R_ph = np.zeros(timestamps.size, dtype=np.float64)
    T_ph = np.zeros(timestamps.size, dtype=np.float64)
    R = rg.randn() * R_sigma + R_mean
    t0 = 0
    for iph, t in enumerate(timestamps):
        # each cycle starts with a new photon timestamp `t`
        # excitation time is `t`, emission time is `t + nanotime`
        delta_t = t - t0  # current time - previous emission time
        if delta_t < 0:
            # avoid negative delta_t possible when when two photons have
            # the same macrotime
            delta_t = 0
            t = t0
        # computed `R_prev`, the D-A distance at the excitation time
        N = rg.randn()
        R = ou_single_step_cy(R, delta_t, N, R_mean, R_sigma, τ_relax)
        nanotime = 0
        # loop through D-A diffusion steps with a fixed time-step δt
        # until D de-excitation by photon emission or energy transfer to A
        while True:
            # with current `R`, we compute the transition rates
            k_ET = k_D * (R0 / R)**6
            k_emission = k_ET + k_D
            # random trial to emit in the current time bin δt
            d_prob_ph_em = k_emission * δt  # prob. of emission in δt
            p = rg.rand()
            if d_prob_ph_em >= p:
                break   # break out of the loop when the photon is emitted
            nanotime += δt
            N = rg.randn()
            R = ou_single_step_cy(R, δt, N, R_mean, R_sigma, τ_relax)

        # photon emitted, let's decide if it is from D or A
        p_DA = p / d_prob_ph_em  # equivalent to rand(), but faster
        prob_A_em = k_ET / k_emission
        if prob_A_em >= p_DA:
            A_ph[iph] = True
        # time of D de-excitation by photon emission or energy transfer to A
        t0 = t + nanotime
        # save D-A distance at emission time
        R_ph[iph] = R
        # save time of emission relative to the excitation time `t`
        T_ph[iph] = nanotime
    return A_ph, R_ph, T_ph


def sim_DA_from_timestamps_p(timestamps, δt_max, k_D, R0, R_mean, R_sigma,
                             τ_relax, rg, α=0.05, ndt=10):
    """
    Recoloring using adaptive δt and no random number caching.

    The adaptive δt is chosen so that k_emission * δt < 0.05,
    so that k_emission * δt is a good approximation for the exponential CDF
    (1 - exp(-k_emission δt)).
    """
    if τ_relax < ndt * δt_max:
        δt_max = τ_relax / ndt
        print(f'WARNING: Reducing δt_max to {δt_max:g}')
    A_ph = np.zeros(timestamps.size, dtype=bool)
    R_ph = np.zeros(timestamps.size, dtype=np.float64)
    T_ph = np.zeros(timestamps.size, dtype=np.float64)
    R = rg.randn() * R_sigma + R_mean
    t0 = 0
    for iph, t in enumerate(timestamps):
        # each cycle starts with a new photon timestamp `t`
        # excitation time is `t`, emission time is `t + nanotime`
        delta_t = t - t0
        if delta_t < 0:
            # avoid negative delta_t possible when when two photons have
            # the same macrotime
            delta_t = 0
            t = t0
        # computed `R_prev`, the D-A distance at the excitation time
        N = rg.randn()
        R = ou_single_step_cy(R, delta_t, N, R_mean, R_sigma, τ_relax)
        nanotime = 0
        # loop through D-A diffusion steps with a time-step δt
        # until D de-excitation by photon emission or energy transfer to A
        while True:
            k_ET = k_D * (R0 / R)**6
            k_emission = k_ET + k_D
            δt = min(α / k_emission, δt_max)
            d_prob_ph_em = k_emission * δt  # prob. of emission in δt
            p = rg.rand()
            if d_prob_ph_em >= p:
                break   # break out of the loop when the photon is emitted
            nanotime += δt
            N = rg.randn()
            R = ou_single_step_cy(R, δt, N, R_mean, R_sigma, τ_relax)

        # photon emitted, let's decide if it is from D or A
        p_DA = p / d_prob_ph_em  # equivalent to rand(), but faster
        prob_A_em = k_ET / k_emission
        if prob_A_em >= p_DA:
            A_ph[iph] = True
        # time of D de-excitation by photon emission or energy transfer to A
        t0 = t + nanotime
        # save D-A distance at emission time
        R_ph[iph] = R
        # save time of emission relative to the excitation time `t`
        T_ph[iph] = nanotime
    return A_ph, R_ph, T_ph


def sim_DA_from_timestamps_p2(timestamps, δt, k_D, R0, R_mean, R_sigma,
                              τ_relax, rg, α=0.05, ndt=10):
    """
    Recoloring using fixed δt, emission are using exponential CDF and
    no random number caching.

    The assumption here is that since τ_relax >> δt, distance and FRET
    will not change during δt, so the transition probability can be computed
    from the exponential CDF.
    """
    A_ph = np.zeros(timestamps.size, dtype=bool)
    R_ph = np.zeros(timestamps.size, dtype=np.float64)
    T_ph = np.zeros(timestamps.size, dtype=np.float64)
    if τ_relax < ndt * δt:
        δt = τ_relax / ndt
        print(f'WARNING: Reducing δt to {δt:g}')
    R = rg.randn() * R_sigma + R_mean
    t0 = 0
    for iph, t in enumerate(timestamps):
        # each cycle starts with a new photon timestamp `t`
        # excitation time is `t`, emission time is `t + nanotime`
        delta_t = t - t0
        if delta_t < 0:
            # avoid negative delta_t possible when when two photons have
            # the same macrotime
            delta_t = 0
            t = t0
        # `R` here is the D-A distance at the excitation time
        N = rg.randn()
        R = ou_single_step_cy(R, delta_t, N, R_mean, R_sigma, τ_relax)
        nanotime = 0
        # loop through D-A diffusion steps with a fixed time-step δt
        # until D de-excitation by photon emission or energy transfer to A
        while True:
            k_ET = k_D * (R0 / R)**6
            k_emission = k_ET + k_D
            d_prob_ph_em = k_emission * δt  # prob. of emission in δt
            if d_prob_ph_em > α:
                d_prob_ph_em = 1 - exp(-d_prob_ph_em)
            p = rg.rand()
            if d_prob_ph_em >= p:
                break   # break out of the loop when the photon is emitted
            nanotime += δt
            # Update R following the OU process
            N = rg.randn()
            R = ou_single_step_cy(R, δt, N, R_mean, R_sigma, τ_relax)

        # photon emitted, let's decide if it is from D or A
        p_DA = p / d_prob_ph_em  # equivalent to rand(), but faster
        prob_A_em = k_ET / k_emission
        if prob_A_em >= p_DA:
            A_ph[iph] = True
        # time of D de-excitation by photon emission or energy transfer to A
        t0 = t + nanotime
        # save D-A distance at emission time
        R_ph[iph] = R
        # save time of emission relative to the excitation time `t`
        T_ph[iph] = nanotime
    return A_ph, R_ph, T_ph


def sim_DA_from_timestamps2(timestamps, dt, k_D, R0, R_mean, R_sigma,
                            τ_relax, rg, chunk_size=1000):
    """
    Recoloring using a fixed δt and with random number caching
    """
    R = rg.randn() * R_sigma + R_mean
    t0 = 0
    nanotime = 0
    # Array flagging photons as A (1) or D (0) emitted
    A_ph = np.zeros(timestamps.size, dtype=bool)
    # Instantaneous D-A distance at D de-excitation time
    R_ph = np.zeros(timestamps.size, dtype=np.float64)
    # Time of D de-excitation relative to the last timestamp
    T_ph = np.zeros(timestamps.size, dtype=np.float64)
    iN = chunk_size - 1
    for iph, t in enumerate(timestamps):
        # each cycle starts with a new photon timestamp `t`
        # excitation time is `t`, emission time is `t + nanotime`
        delta_t = t - t0
        if delta_t < 0:
            # avoid negative delta_t possible when when two photons have
            # the same macrotime
            delta_t = 0
            t = t0
        # Compute the D-A distance at the "excitation time"
        iN += 1
        if iN == chunk_size:
            #print(f'Outer rand iph:{iph}, iN:{iN}', flush=True)
            Na = memoryview(rg.randn(chunk_size))
            Pa = memoryview(rg.rand(chunk_size))
            iN = 0
        N = Na[iN]
        p = Pa[iN]
        R = ou_single_step_cy(R, delta_t, N, R_mean, R_sigma, τ_relax)
        nanotime = 0
        # loop through D-A diffusion steps with a fixed time-step dt
        # until D de-excitation by photon emission or energy transfer to A
        while True:
            k_ET = k_D * (R0 / R)**6
            k_emission = k_ET + k_D
            d_prob_ph_em = k_emission * dt  # prob. of emission in dt
            if d_prob_ph_em >= p:
                break   # break out of the loop when the photon is emitted
            nanotime += dt
            iN += 1
            if iN == chunk_size:
                Na = memoryview(rg.randn(chunk_size))
                Pa = memoryview(rg.rand(chunk_size))
                iN = 0
            N = Na[iN]
            p = Pa[iN]
            R = ou_single_step_cy(R, dt, N, R_mean, R_sigma, τ_relax)

        # photon emitted, let's decide if it is from D or A
        p_DA = p / d_prob_ph_em  # equivalent to rand(), but faster
        prob_A_em = k_ET / (k_ET + k_D)
        if prob_A_em >= p_DA:
            A_ph[iph] = True
        # time of D de-excitation by photon emission or energy transfer to A
        t0 = t + nanotime
        # save D-A distance at emission time
        R_ph[iph] = R
        # save time of emission relative to the excitation time `t`
        T_ph[iph] = nanotime
    return A_ph, R_ph, T_ph


def sim_DA_from_timestamps2_p(timestamps, dt_max, k_D, R0, R_mean, R_sigma,
                              tau_relax, rg, chunk_size=1000,
                              alpha=0.05, ndt=10):
    """
    Recoloring using adaptive dt and with random number caching

    The adaptive dt is chosen so that `k_emission * dt < 0.05`,
    so that k_emission * dt is a good approximation for the exponential CDF
    (1 - exp(-k_emission dt)).
    """
    if tau_relax < ndt * dt_max:
        dt_max = tau_relax / ndt
        print(f'WARNING: Reducing dt_max to {dt_max:g} '
              f'[tau_relax = {tau_relax}]')
    R = rg.randn() * R_sigma + R_mean
    t0 = 0
    nanotime = 0
    # Array flagging photons as A (1) or D (0) emitted
    A_ph = np.zeros(timestamps.size, dtype=bool)
    # Instantaneous D-A distance at D de-excitation time
    R_ph = np.zeros(timestamps.size, dtype=np.float64)
    # Time of D de-excitation relative to the last timestamp
    T_ph = np.zeros(timestamps.size, dtype=np.float64)
    iN = chunk_size - 1  # value to get the first chunk of random numbers
    for iph, t in enumerate(timestamps):
        # each cycle starts with a new photon timestamp `t`
        # excitation time is `t`, emission time is `t + nanotime`
        delta_t = t - t0
        if delta_t < 0:
            # avoid negative delta_t possible when when two photons have
            # the same macrotime
            delta_t = 0
            t = t0
        # Compute the D-A distance at the "excitation time"
        iN += 1
        if iN == chunk_size:
            Na = memoryview(rg.randn(chunk_size))
            Pa = memoryview(rg.rand(chunk_size))
            iN = 0
        N = Na[iN]
        p = Pa[iN]
        R = ou_single_step_cy(R, delta_t, N, R_mean, R_sigma, tau_relax)
        nanotime = 0
        # loop through D-A diffusion steps with a fixed time-step dt
        # until D de-excitation by photon emission or energy transfer to A
        while True:
            k_ET = k_D * (R0 / R)**6
            k_emission = k_ET + k_D
            dt = min(alpha / k_emission, dt_max)
            d_prob_ph_em = k_emission * dt  # prob. of emission in dt
            if d_prob_ph_em >= p:
                break   # break out of the loop when the photon is emitted
            nanotime += dt
            iN += 1
            if iN == chunk_size:
                Na = memoryview(rg.randn(chunk_size))
                Pa = memoryview(rg.rand(chunk_size))
                iN = 0
            N = Na[iN]
            p = Pa[iN]
            # Update R following the OU process
            R = ou_single_step_cy(R, dt, N, R_mean, R_sigma, tau_relax)
        # photon emitted, let's decide if it is from D or A
        p_DA = p / d_prob_ph_em  # equivalent to rand(), but faster
        prob_A_em = k_ET / (k_ET + k_D)
        if prob_A_em >= p_DA:
            A_ph[iph] = True
        # time of D de-excitation by photon emission or energy transfer to A
        t0 = t + nanotime
        # save D-A distance at emission time
        R_ph[iph] = R
        # save time of emission relative to the excitation time `t`
        T_ph[iph] = nanotime
    return A_ph, R_ph, T_ph


#
# - FOLLOWING FUNCTIONS ARE NOT USED FOR RECOLORING - - - - - - - - - - -
#


@numba.jit(nopython=True)
def ou_process_core_numba(X, N, delta_t, ou_tau, ou_sigma):
    """Low-level function computing values of the OU process
    """
    for i in range(1, X.size):
        dt = delta_t[i - 1]
        dt_over_tau = dt / ou_tau
        relax = np.exp(-dt_over_tau)
        diffuse = ou_sigma * np.sqrt(1 - relax**2)
        X[i] = X[i - 1] * relax + diffuse * N[i]


def ou_process(delta_t, N, ou_mean, ou_sigma, ou_tau, x0=None):
    """Compute an Ornstein–Uhlenbeck (OU) process from a series of delta_t.

    Arguments:
        delta_t (array): intervals between time points where the OU process
            will be evaluated
        N (array): array of white Gaussian noise with sigma=1 of size
            size(delta_t) + 1.
        ou_mean (float): mean of the OU process
        ou_sigma (float): standard deviation of the relaxed OU process
        ou_tau (float): relaxation time of the OU process
        x0 (float): initial value for the OU process. If None,
            the initial value is `N[0] * ou_sigma`. If not None,
            `N[0]` is not used.
    Returns:
        Array of OU process values. Output size is `size(delta_t) + 1`
    """
    X = np.zeros(delta_t.size + 1, dtype=float)
    if x0 is None:
        X[0] = N[0] * ou_sigma
    else:
        X[0] = x0
    ou_process_core_numba(X, N, delta_t, ou_tau, ou_sigma)
    X += ou_mean
    return X


def E_sim_gauss_R_burst(burst_size_series, R0, R_mean, τ_relax=0,
                        R_sigma=0, oversample=100, dithering_sigma=0.02):
    """Simulate a E for each burst using a fixed D-A distance per burst.

    Distances are drawn from a Gaussian distribution.

    Arguments:
        burst_size_series (pandas.Series): series of burst sizes
    """
    size_counts = burst_size_series.value_counts()
    E_sim = []
    for burst_size, counts in size_counts.iteritems():
        num_bursts = oversample * counts
        R = np.random.randn(num_bursts) * R_sigma + R_mean
        assert (R >= 0).all()
        E = fret.E_from_dist(R, R0)
        assert (E >= 0).all() and (E <= 1).all()
        na_sim = np.random.binomial(n=burst_size, p=E, size=num_bursts)
        E_sim.append(na_sim / burst_size)
    E_sim = np.hstack(E_sim)
    assert E_sim.size == size_counts.sum() * oversample
    E_sim += np.random.randn(E_sim.size) * dithering_sigma
    return E_sim


def E_sim_gauss_R_ph(burst_size_series, R0, R_mean, τ_relax=0,
                     R_sigma=0, oversample=100, dithering_sigma=0.02):
    """Simulate E for each burst using a new Gaussian D-A distance per photon

    - R drawn from Gaussian distribution for each photon in each burst
    - Group by burst size

    Arguments:
        burst_size_series (pandas.Series): series of burst sizes
    """
    size_counts = burst_size_series.value_counts()
    E_sim = []
    for burst_size, counts in size_counts.iteritems():
        num_bursts = oversample * counts
        R = np.random.randn(num_bursts, burst_size) * R_sigma + R_mean
        assert (R >= 0).all()
        E = fret.E_from_dist(R, R0)
        assert (E >= 0).all() and (E <= 1).all()
        na_sim = np.random.binomial(n=1, p=E).sum(axis=1)
        E_sim.append(na_sim / burst_size)
    E_sim = np.hstack(E_sim)
    assert E_sim.size == size_counts.sum() * oversample
    E_sim += np.random.randn(E_sim.size) * dithering_sigma
    return E_sim


def E_sim(E_ph, oversample=100, dithering_sigma=0.02):
    """Simulate E for each burst using the passed `E_ph` for each photon.
    """
    num_bursts = oversample
    E_sim = []
    for ib, E in E_ph.groupby('burst'):
        burst_size = E.shape[0]
        assert (E >= 0).all() and (E <= 1).all()
        na_sim = (np.random.binomial(n=1, p=E, size=(num_bursts, burst_size))
                  .sum(axis=1))
        E_sim.append(na_sim / burst_size)
    E_sim = np.hstack(E_sim)
    E_sim += np.random.randn(E_sim.size) * dithering_sigma
    return E_sim



import numpy as np
import pandas as pd


def correct_E_gamma_leak_dir(Eraw, gamma=1, leakage=0, dir_ex_t=0):
    """Compute corrected FRET efficiency from proximity ratio `Eraw`.

    This function is the inverse of :func:`uncorrect_E_gamma_leak_dir`.

    Arguments:
        Eraw (float or array): proximity ratio (only background correction,
            no gamma, leakage or direct excitation)
        gamma (float): gamma factor
        leakage (float): leakage coefficient
        dir_ex_t (float): coefficient expressing the direct excitation as
            n_dir = dir_ex_t * (na + gamma*nd). In terms of physical
            parameters it is the ratio of acceptor over donor absorption
            cross-sections at the donor-excitation wavelength.

    Returns
        Corrected FRET effciency
    """
    if not hasattr(Eraw, '__array__'):
        Eraw = np.asarray(Eraw)
    return ((Eraw * (leakage + dir_ex_t * gamma + 1) - leakage - dir_ex_t * gamma)
            / (Eraw * (leakage - gamma + 1) - leakage + gamma))


def uncorrect_E_gamma_leak_dir(E, gamma=1, leakage=0, dir_ex_t=0):
    """Compute proximity ratio from corrected FRET efficiency `E`.

    This function is the inverse of :func:`correct_E_gamma_leak_dir`.

    Arguments:
        E (float or array): corrected FRET efficiency
        gamma (float): gamma factor
        leakage (float): leakage coefficient
        dir_ex_t (float): direct excitation coefficient expressed as
            n_dir = dir_ex_t * (na + gamma*nd). In terms of physical
            parameters it is the ratio of absorption cross-section at
            donor-excitation wavelengths of acceptor over donor.

    Returns
        Proximity ratio (reverses gamma, leakage and direct excitation)
    """
    if not hasattr(E, '__array__'):
        E = np.asarray(E)
    return ((E * (gamma - leakage) + leakage + dir_ex_t * gamma)
            / (E * (gamma - leakage - 1) + leakage + dir_ex_t * gamma + 1))


def E_from_dist(r, R0):
    """Return E computed from D-A distance and R0
    """
    if not hasattr(r, '__array__'):
        r = np.asarray(r)
    E = 1 / (1 + (r / R0)**6)
    if r.ndim > 0:
        E[r < 0] = 1
    elif r < 0:
        E = 0
    return E


def dist_from_E(E, R0):
    """Return the D-A distance for a give E and R0
    """
    if not hasattr(E, '__array__'):
        E = np.asarray(E)
    return R0 * (1 / E - 1)**(1 / 6)


def mean_E_from_gauss_PoR(R_mean, R_sigma, R0):
    """Mean E from integration of a Gaussian distribution of distances (PoR)
    """
    def gaussian(x, mu, sig):
        return np.exp(-np.power(x - mu, 2.) / (2 * np.power(sig, 2.)))
    R_axis = np.arange(0, R_mean + 5 * R_sigma, R_sigma / 200)
    R_pdf = gaussian(R_axis, R_mean, R_sigma)
    R_pdf /= np.trapz(R_pdf, R_axis)
    EoR = E_from_dist(R_axis, R0)
    E_mean_integral = np.trapz(EoR * R_pdf, R_axis)
    return E_mean_integral


def calc_E_burst(burstsph):
    """Calculate E from photon-data (photon timestamp and stream)

    Arguments:
        burstsph (DataFrame): photon data DataFrame with at least two
            columns: 'timestamp' and 'stream'. The index needs to have
            two levels: ('burst', 'ph').

    Returns:
        Array of E values, one for each burst.
    """
    bursts = pd.DataFrame(burstsph.groupby('burst').size(),
                          columns=['size_raw'])
    bursts['istart'] = np.hstack([[0], np.cumsum(bursts.size_raw)[:-1]])
    bursts['istop'] = np.cumsum(bursts.size_raw.values)
    A_em = burstsph.stream == 'DexAem'
    D_ex = (burstsph.stream == 'DexDem') | (burstsph.stream == 'DexAem')
    E = [A_em[istart:istop].sum() / D_ex[istart:istop].sum()
         for istart, istop, bsize in
         zip(bursts.istart, bursts.istop, bursts.size_raw)]
    return E
